# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union, Iterable
from typing_extensions import Literal

import httpx

from ..types import aggregation_list_params, aggregation_create_params, aggregation_update_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import (
    maybe_transform,
    async_maybe_transform,
)
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ..pagination import SyncCursor, AsyncCursor
from .._base_client import AsyncPaginator, make_request_options
from ..types.aggregation_response import AggregationResponse

__all__ = ["AggregationsResource", "AsyncAggregationsResource"]


class AggregationsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AggregationsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#accessing-raw-response-data-eg-headers
        """
        return AggregationsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AggregationsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#with_streaming_response
        """
        return AggregationsResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        org_id: str | None = None,
        aggregation: Literal["SUM", "MIN", "MAX", "COUNT", "LATEST", "MEAN", "UNIQUE", "CUSTOM_SQL"],
        meter_id: str,
        name: str,
        quantity_per_unit: float,
        rounding: Literal["UP", "DOWN", "NEAREST", "NONE"],
        target_field: str,
        unit: str,
        accounting_product_id: str | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        custom_sql: str | NotGiven = NOT_GIVEN,
        default_value: float | NotGiven = NOT_GIVEN,
        segmented_fields: List[str] | NotGiven = NOT_GIVEN,
        segments: Iterable[Dict[str, str]] | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Create a new Aggregation.

        Args:
          aggregation: Specifies the computation method applied to usage data collected in
              `targetField`. Aggregation unit value depends on the **Category** configured for
              the selected targetField.

              Enum:

              - **SUM**. Adds the values. Can be applied to a **Measure**, **Income**, or
                **Cost** `targetField`.

              - **MIN**. Uses the minimum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **MAX**. Uses the maximum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **COUNT**. Counts the number of values. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`.

              - **LATEST**. Uses the most recent value. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`. Note: Based on the timestamp (`ts`)
                value of usage data measurement submissions. If using this method, please
                ensure _distinct_ `ts` values are used for usage data measurment submissions.

              - **MEAN**. Uses the arithmetic mean of the values. Can be applied to a
                **Measure**, **Income**, or **Cost** `targetField`.

              - **UNIQUE**. Uses unique values and returns a count of the number of unique
                values. Can be applied to a **Metadata** `targetField`.

          meter_id: The UUID of the Meter used as the source of usage data for the Aggregation.

              Each Aggregation is a child of a Meter, so the Meter must be selected.

          name: Descriptive name for the Aggregation.

          quantity_per_unit: Defines how much of a quantity equates to 1 unit. Used when setting the price
              per unit for billing purposes - if charging for kilobytes per second (KiBy/s) at
              rate of $0.25 per 500 KiBy/s, then set quantityPerUnit to 500 and price Plan at
              $0.25 per unit.

              **Note:** If `quantityPerUnit` is set to a value other than one, `rounding` is
              typically set to `"UP"`.

          rounding: Specifies how you want to deal with non-integer, fractional number Aggregation
              values.

              **NOTES:**

              - **NEAREST** rounds to the nearest half: 5.1 is rounded to 5, and 3.5 is
                rounded to 4.
              - Also used in combination with `quantityPerUnit`. Rounds the number of units
                after `quantityPerUnit` is applied. If you set `quantityPerUnit` to a value
                other than one, you would typically set Rounding to **UP**. For example,
                suppose you charge by kilobytes per second (KiBy/s), set `quantityPerUnit` =
                500, and set charge rate at $0.25 per unit used. If your customer used 48,900
                KiBy/s in a billing period, the charge would be 48,900 / 500 = 97.8 rounded up
                to 98 \\** 0.25 = $2.45.

              Enum: ???UP??? ???DOWN??? ???NEAREST??? ???NONE???

          target_field: `Code` of the target `dataField` or `derivedField` on the Meter used as the
              basis for the Aggregation.

          unit: User defined label for units shown for Bill line items, indicating to your
              customers what they are being charged for.

          accounting_product_id: Optional Product ID this Aggregation should be attributed to for accounting
              purposes

          code: Code of the new Aggregation. A unique short code to identify the Aggregation.

          custom_sql

          default_value: Aggregation value used when no usage data is available to be aggregated.
              _(Optional)_.

              **Note:** Set to 0, if you expect to reference the Aggregation in a Compound
              Aggregation. This ensures that any null values are passed in correctly to the
              Compound Aggregation calculation with a value = 0.

          segmented_fields: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segments`.

              Enter the `Codes` of the fields in the target Meter to use for segmentation
              purposes.

              String `dataFields` on the target Meter can be segmented. Any string
              `derivedFields` on the target Meter, such as one that concatenates two string
              `dataFields`, can also be segmented.

          segments: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segmentedFields`.

              Enter the values that are to be used as the segments, read from the fields in
              the meter pointed at by `segmentedFields`.

              Note that you can use _wildcards_ or _defaults_ when setting up segment values.
              For more details on how to do this with an example, see
              [Using Wildcards - API Calls](https://www.m3ter.com/docs/guides/setting-up-usage-data-meters-and-aggregations/segmented-aggregations#using-wildcards---api-calls)
              in our main User Docs.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return self._post(
            f"/organizations/{org_id}/aggregations",
            body=maybe_transform(
                {
                    "aggregation": aggregation,
                    "meter_id": meter_id,
                    "name": name,
                    "quantity_per_unit": quantity_per_unit,
                    "rounding": rounding,
                    "target_field": target_field,
                    "unit": unit,
                    "accounting_product_id": accounting_product_id,
                    "code": code,
                    "custom_fields": custom_fields,
                    "custom_sql": custom_sql,
                    "default_value": default_value,
                    "segmented_fields": segmented_fields,
                    "segments": segments,
                    "version": version,
                },
                aggregation_create_params.AggregationCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )

    def retrieve(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Retrieve the Aggregation with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._get(
            f"/organizations/{org_id}/aggregations/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )

    def update(
        self,
        id: str,
        *,
        org_id: str | None = None,
        aggregation: Literal["SUM", "MIN", "MAX", "COUNT", "LATEST", "MEAN", "UNIQUE", "CUSTOM_SQL"],
        meter_id: str,
        name: str,
        quantity_per_unit: float,
        rounding: Literal["UP", "DOWN", "NEAREST", "NONE"],
        target_field: str,
        unit: str,
        accounting_product_id: str | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        custom_sql: str | NotGiven = NOT_GIVEN,
        default_value: float | NotGiven = NOT_GIVEN,
        segmented_fields: List[str] | NotGiven = NOT_GIVEN,
        segments: Iterable[Dict[str, str]] | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Update the Aggregation with the given UUID.

        **Note:** If you have created Custom Fields for an Aggregation, when you use
        this endpoint to update the Aggregation use the `customFields` parameter to
        preserve those Custom Fields. If you omit them from the update request, they
        will be lost.

        Args:
          aggregation: Specifies the computation method applied to usage data collected in
              `targetField`. Aggregation unit value depends on the **Category** configured for
              the selected targetField.

              Enum:

              - **SUM**. Adds the values. Can be applied to a **Measure**, **Income**, or
                **Cost** `targetField`.

              - **MIN**. Uses the minimum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **MAX**. Uses the maximum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **COUNT**. Counts the number of values. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`.

              - **LATEST**. Uses the most recent value. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`. Note: Based on the timestamp (`ts`)
                value of usage data measurement submissions. If using this method, please
                ensure _distinct_ `ts` values are used for usage data measurment submissions.

              - **MEAN**. Uses the arithmetic mean of the values. Can be applied to a
                **Measure**, **Income**, or **Cost** `targetField`.

              - **UNIQUE**. Uses unique values and returns a count of the number of unique
                values. Can be applied to a **Metadata** `targetField`.

          meter_id: The UUID of the Meter used as the source of usage data for the Aggregation.

              Each Aggregation is a child of a Meter, so the Meter must be selected.

          name: Descriptive name for the Aggregation.

          quantity_per_unit: Defines how much of a quantity equates to 1 unit. Used when setting the price
              per unit for billing purposes - if charging for kilobytes per second (KiBy/s) at
              rate of $0.25 per 500 KiBy/s, then set quantityPerUnit to 500 and price Plan at
              $0.25 per unit.

              **Note:** If `quantityPerUnit` is set to a value other than one, `rounding` is
              typically set to `"UP"`.

          rounding: Specifies how you want to deal with non-integer, fractional number Aggregation
              values.

              **NOTES:**

              - **NEAREST** rounds to the nearest half: 5.1 is rounded to 5, and 3.5 is
                rounded to 4.
              - Also used in combination with `quantityPerUnit`. Rounds the number of units
                after `quantityPerUnit` is applied. If you set `quantityPerUnit` to a value
                other than one, you would typically set Rounding to **UP**. For example,
                suppose you charge by kilobytes per second (KiBy/s), set `quantityPerUnit` =
                500, and set charge rate at $0.25 per unit used. If your customer used 48,900
                KiBy/s in a billing period, the charge would be 48,900 / 500 = 97.8 rounded up
                to 98 \\** 0.25 = $2.45.

              Enum: ???UP??? ???DOWN??? ???NEAREST??? ???NONE???

          target_field: `Code` of the target `dataField` or `derivedField` on the Meter used as the
              basis for the Aggregation.

          unit: User defined label for units shown for Bill line items, indicating to your
              customers what they are being charged for.

          accounting_product_id: Optional Product ID this Aggregation should be attributed to for accounting
              purposes

          code: Code of the new Aggregation. A unique short code to identify the Aggregation.

          custom_sql

          default_value: Aggregation value used when no usage data is available to be aggregated.
              _(Optional)_.

              **Note:** Set to 0, if you expect to reference the Aggregation in a Compound
              Aggregation. This ensures that any null values are passed in correctly to the
              Compound Aggregation calculation with a value = 0.

          segmented_fields: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segments`.

              Enter the `Codes` of the fields in the target Meter to use for segmentation
              purposes.

              String `dataFields` on the target Meter can be segmented. Any string
              `derivedFields` on the target Meter, such as one that concatenates two string
              `dataFields`, can also be segmented.

          segments: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segmentedFields`.

              Enter the values that are to be used as the segments, read from the fields in
              the meter pointed at by `segmentedFields`.

              Note that you can use _wildcards_ or _defaults_ when setting up segment values.
              For more details on how to do this with an example, see
              [Using Wildcards - API Calls](https://www.m3ter.com/docs/guides/setting-up-usage-data-meters-and-aggregations/segmented-aggregations#using-wildcards---api-calls)
              in our main User Docs.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._put(
            f"/organizations/{org_id}/aggregations/{id}",
            body=maybe_transform(
                {
                    "aggregation": aggregation,
                    "meter_id": meter_id,
                    "name": name,
                    "quantity_per_unit": quantity_per_unit,
                    "rounding": rounding,
                    "target_field": target_field,
                    "unit": unit,
                    "accounting_product_id": accounting_product_id,
                    "code": code,
                    "custom_fields": custom_fields,
                    "custom_sql": custom_sql,
                    "default_value": default_value,
                    "segmented_fields": segmented_fields,
                    "segments": segments,
                    "version": version,
                },
                aggregation_update_params.AggregationUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )

    def list(
        self,
        *,
        org_id: str | None = None,
        codes: List[str] | NotGiven = NOT_GIVEN,
        ids: List[str] | NotGiven = NOT_GIVEN,
        next_token: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        product_id: List[str] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncCursor[AggregationResponse]:
        """
        Retrieve a list of Aggregations that can be filtered by Product, Aggregation ID,
        or Code.

        Args:
          codes: List of Aggregation codes to retrieve. These are unique short codes to identify
              each Aggregation.

          ids: List of Aggregation IDs to retrieve.

          next_token: `nextToken` for multi-page retrievals.

          page_size: Number of Aggregations to retrieve per page.

          product_id: The UUIDs of the Products to retrieve Aggregations for.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return self._get_api_list(
            f"/organizations/{org_id}/aggregations",
            page=SyncCursor[AggregationResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "codes": codes,
                        "ids": ids,
                        "next_token": next_token,
                        "page_size": page_size,
                        "product_id": product_id,
                    },
                    aggregation_list_params.AggregationListParams,
                ),
            ),
            model=AggregationResponse,
        )

    def delete(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Delete the Aggregation with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._delete(
            f"/organizations/{org_id}/aggregations/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )


class AsyncAggregationsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncAggregationsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#accessing-raw-response-data-eg-headers
        """
        return AsyncAggregationsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncAggregationsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#with_streaming_response
        """
        return AsyncAggregationsResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        org_id: str | None = None,
        aggregation: Literal["SUM", "MIN", "MAX", "COUNT", "LATEST", "MEAN", "UNIQUE", "CUSTOM_SQL"],
        meter_id: str,
        name: str,
        quantity_per_unit: float,
        rounding: Literal["UP", "DOWN", "NEAREST", "NONE"],
        target_field: str,
        unit: str,
        accounting_product_id: str | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        custom_sql: str | NotGiven = NOT_GIVEN,
        default_value: float | NotGiven = NOT_GIVEN,
        segmented_fields: List[str] | NotGiven = NOT_GIVEN,
        segments: Iterable[Dict[str, str]] | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Create a new Aggregation.

        Args:
          aggregation: Specifies the computation method applied to usage data collected in
              `targetField`. Aggregation unit value depends on the **Category** configured for
              the selected targetField.

              Enum:

              - **SUM**. Adds the values. Can be applied to a **Measure**, **Income**, or
                **Cost** `targetField`.

              - **MIN**. Uses the minimum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **MAX**. Uses the maximum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **COUNT**. Counts the number of values. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`.

              - **LATEST**. Uses the most recent value. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`. Note: Based on the timestamp (`ts`)
                value of usage data measurement submissions. If using this method, please
                ensure _distinct_ `ts` values are used for usage data measurment submissions.

              - **MEAN**. Uses the arithmetic mean of the values. Can be applied to a
                **Measure**, **Income**, or **Cost** `targetField`.

              - **UNIQUE**. Uses unique values and returns a count of the number of unique
                values. Can be applied to a **Metadata** `targetField`.

          meter_id: The UUID of the Meter used as the source of usage data for the Aggregation.

              Each Aggregation is a child of a Meter, so the Meter must be selected.

          name: Descriptive name for the Aggregation.

          quantity_per_unit: Defines how much of a quantity equates to 1 unit. Used when setting the price
              per unit for billing purposes - if charging for kilobytes per second (KiBy/s) at
              rate of $0.25 per 500 KiBy/s, then set quantityPerUnit to 500 and price Plan at
              $0.25 per unit.

              **Note:** If `quantityPerUnit` is set to a value other than one, `rounding` is
              typically set to `"UP"`.

          rounding: Specifies how you want to deal with non-integer, fractional number Aggregation
              values.

              **NOTES:**

              - **NEAREST** rounds to the nearest half: 5.1 is rounded to 5, and 3.5 is
                rounded to 4.
              - Also used in combination with `quantityPerUnit`. Rounds the number of units
                after `quantityPerUnit` is applied. If you set `quantityPerUnit` to a value
                other than one, you would typically set Rounding to **UP**. For example,
                suppose you charge by kilobytes per second (KiBy/s), set `quantityPerUnit` =
                500, and set charge rate at $0.25 per unit used. If your customer used 48,900
                KiBy/s in a billing period, the charge would be 48,900 / 500 = 97.8 rounded up
                to 98 \\** 0.25 = $2.45.

              Enum: ???UP??? ???DOWN??? ???NEAREST??? ???NONE???

          target_field: `Code` of the target `dataField` or `derivedField` on the Meter used as the
              basis for the Aggregation.

          unit: User defined label for units shown for Bill line items, indicating to your
              customers what they are being charged for.

          accounting_product_id: Optional Product ID this Aggregation should be attributed to for accounting
              purposes

          code: Code of the new Aggregation. A unique short code to identify the Aggregation.

          custom_sql

          default_value: Aggregation value used when no usage data is available to be aggregated.
              _(Optional)_.

              **Note:** Set to 0, if you expect to reference the Aggregation in a Compound
              Aggregation. This ensures that any null values are passed in correctly to the
              Compound Aggregation calculation with a value = 0.

          segmented_fields: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segments`.

              Enter the `Codes` of the fields in the target Meter to use for segmentation
              purposes.

              String `dataFields` on the target Meter can be segmented. Any string
              `derivedFields` on the target Meter, such as one that concatenates two string
              `dataFields`, can also be segmented.

          segments: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segmentedFields`.

              Enter the values that are to be used as the segments, read from the fields in
              the meter pointed at by `segmentedFields`.

              Note that you can use _wildcards_ or _defaults_ when setting up segment values.
              For more details on how to do this with an example, see
              [Using Wildcards - API Calls](https://www.m3ter.com/docs/guides/setting-up-usage-data-meters-and-aggregations/segmented-aggregations#using-wildcards---api-calls)
              in our main User Docs.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return await self._post(
            f"/organizations/{org_id}/aggregations",
            body=await async_maybe_transform(
                {
                    "aggregation": aggregation,
                    "meter_id": meter_id,
                    "name": name,
                    "quantity_per_unit": quantity_per_unit,
                    "rounding": rounding,
                    "target_field": target_field,
                    "unit": unit,
                    "accounting_product_id": accounting_product_id,
                    "code": code,
                    "custom_fields": custom_fields,
                    "custom_sql": custom_sql,
                    "default_value": default_value,
                    "segmented_fields": segmented_fields,
                    "segments": segments,
                    "version": version,
                },
                aggregation_create_params.AggregationCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )

    async def retrieve(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Retrieve the Aggregation with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._get(
            f"/organizations/{org_id}/aggregations/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )

    async def update(
        self,
        id: str,
        *,
        org_id: str | None = None,
        aggregation: Literal["SUM", "MIN", "MAX", "COUNT", "LATEST", "MEAN", "UNIQUE", "CUSTOM_SQL"],
        meter_id: str,
        name: str,
        quantity_per_unit: float,
        rounding: Literal["UP", "DOWN", "NEAREST", "NONE"],
        target_field: str,
        unit: str,
        accounting_product_id: str | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        custom_sql: str | NotGiven = NOT_GIVEN,
        default_value: float | NotGiven = NOT_GIVEN,
        segmented_fields: List[str] | NotGiven = NOT_GIVEN,
        segments: Iterable[Dict[str, str]] | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Update the Aggregation with the given UUID.

        **Note:** If you have created Custom Fields for an Aggregation, when you use
        this endpoint to update the Aggregation use the `customFields` parameter to
        preserve those Custom Fields. If you omit them from the update request, they
        will be lost.

        Args:
          aggregation: Specifies the computation method applied to usage data collected in
              `targetField`. Aggregation unit value depends on the **Category** configured for
              the selected targetField.

              Enum:

              - **SUM**. Adds the values. Can be applied to a **Measure**, **Income**, or
                **Cost** `targetField`.

              - **MIN**. Uses the minimum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **MAX**. Uses the maximum value. Can be applied to a **Measure**, **Income**,
                or **Cost** `targetField`.

              - **COUNT**. Counts the number of values. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`.

              - **LATEST**. Uses the most recent value. Can be applied to a **Measure**,
                **Income**, or **Cost** `targetField`. Note: Based on the timestamp (`ts`)
                value of usage data measurement submissions. If using this method, please
                ensure _distinct_ `ts` values are used for usage data measurment submissions.

              - **MEAN**. Uses the arithmetic mean of the values. Can be applied to a
                **Measure**, **Income**, or **Cost** `targetField`.

              - **UNIQUE**. Uses unique values and returns a count of the number of unique
                values. Can be applied to a **Metadata** `targetField`.

          meter_id: The UUID of the Meter used as the source of usage data for the Aggregation.

              Each Aggregation is a child of a Meter, so the Meter must be selected.

          name: Descriptive name for the Aggregation.

          quantity_per_unit: Defines how much of a quantity equates to 1 unit. Used when setting the price
              per unit for billing purposes - if charging for kilobytes per second (KiBy/s) at
              rate of $0.25 per 500 KiBy/s, then set quantityPerUnit to 500 and price Plan at
              $0.25 per unit.

              **Note:** If `quantityPerUnit` is set to a value other than one, `rounding` is
              typically set to `"UP"`.

          rounding: Specifies how you want to deal with non-integer, fractional number Aggregation
              values.

              **NOTES:**

              - **NEAREST** rounds to the nearest half: 5.1 is rounded to 5, and 3.5 is
                rounded to 4.
              - Also used in combination with `quantityPerUnit`. Rounds the number of units
                after `quantityPerUnit` is applied. If you set `quantityPerUnit` to a value
                other than one, you would typically set Rounding to **UP**. For example,
                suppose you charge by kilobytes per second (KiBy/s), set `quantityPerUnit` =
                500, and set charge rate at $0.25 per unit used. If your customer used 48,900
                KiBy/s in a billing period, the charge would be 48,900 / 500 = 97.8 rounded up
                to 98 \\** 0.25 = $2.45.

              Enum: ???UP??? ???DOWN??? ???NEAREST??? ???NONE???

          target_field: `Code` of the target `dataField` or `derivedField` on the Meter used as the
              basis for the Aggregation.

          unit: User defined label for units shown for Bill line items, indicating to your
              customers what they are being charged for.

          accounting_product_id: Optional Product ID this Aggregation should be attributed to for accounting
              purposes

          code: Code of the new Aggregation. A unique short code to identify the Aggregation.

          custom_sql

          default_value: Aggregation value used when no usage data is available to be aggregated.
              _(Optional)_.

              **Note:** Set to 0, if you expect to reference the Aggregation in a Compound
              Aggregation. This ensures that any null values are passed in correctly to the
              Compound Aggregation calculation with a value = 0.

          segmented_fields: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segments`.

              Enter the `Codes` of the fields in the target Meter to use for segmentation
              purposes.

              String `dataFields` on the target Meter can be segmented. Any string
              `derivedFields` on the target Meter, such as one that concatenates two string
              `dataFields`, can also be segmented.

          segments: _(Optional)_. Used when creating a segmented Aggregation, which segments the
              usage data collected by a single Meter. Works together with `segmentedFields`.

              Enter the values that are to be used as the segments, read from the fields in
              the meter pointed at by `segmentedFields`.

              Note that you can use _wildcards_ or _defaults_ when setting up segment values.
              For more details on how to do this with an example, see
              [Using Wildcards - API Calls](https://www.m3ter.com/docs/guides/setting-up-usage-data-meters-and-aggregations/segmented-aggregations#using-wildcards---api-calls)
              in our main User Docs.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._put(
            f"/organizations/{org_id}/aggregations/{id}",
            body=await async_maybe_transform(
                {
                    "aggregation": aggregation,
                    "meter_id": meter_id,
                    "name": name,
                    "quantity_per_unit": quantity_per_unit,
                    "rounding": rounding,
                    "target_field": target_field,
                    "unit": unit,
                    "accounting_product_id": accounting_product_id,
                    "code": code,
                    "custom_fields": custom_fields,
                    "custom_sql": custom_sql,
                    "default_value": default_value,
                    "segmented_fields": segmented_fields,
                    "segments": segments,
                    "version": version,
                },
                aggregation_update_params.AggregationUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )

    def list(
        self,
        *,
        org_id: str | None = None,
        codes: List[str] | NotGiven = NOT_GIVEN,
        ids: List[str] | NotGiven = NOT_GIVEN,
        next_token: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        product_id: List[str] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[AggregationResponse, AsyncCursor[AggregationResponse]]:
        """
        Retrieve a list of Aggregations that can be filtered by Product, Aggregation ID,
        or Code.

        Args:
          codes: List of Aggregation codes to retrieve. These are unique short codes to identify
              each Aggregation.

          ids: List of Aggregation IDs to retrieve.

          next_token: `nextToken` for multi-page retrievals.

          page_size: Number of Aggregations to retrieve per page.

          product_id: The UUIDs of the Products to retrieve Aggregations for.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return self._get_api_list(
            f"/organizations/{org_id}/aggregations",
            page=AsyncCursor[AggregationResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "codes": codes,
                        "ids": ids,
                        "next_token": next_token,
                        "page_size": page_size,
                        "product_id": product_id,
                    },
                    aggregation_list_params.AggregationListParams,
                ),
            ),
            model=AggregationResponse,
        )

    async def delete(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AggregationResponse:
        """
        Delete the Aggregation with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._delete(
            f"/organizations/{org_id}/aggregations/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=AggregationResponse,
        )


class AggregationsResourceWithRawResponse:
    def __init__(self, aggregations: AggregationsResource) -> None:
        self._aggregations = aggregations

        self.create = to_raw_response_wrapper(
            aggregations.create,
        )
        self.retrieve = to_raw_response_wrapper(
            aggregations.retrieve,
        )
        self.update = to_raw_response_wrapper(
            aggregations.update,
        )
        self.list = to_raw_response_wrapper(
            aggregations.list,
        )
        self.delete = to_raw_response_wrapper(
            aggregations.delete,
        )


class AsyncAggregationsResourceWithRawResponse:
    def __init__(self, aggregations: AsyncAggregationsResource) -> None:
        self._aggregations = aggregations

        self.create = async_to_raw_response_wrapper(
            aggregations.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            aggregations.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            aggregations.update,
        )
        self.list = async_to_raw_response_wrapper(
            aggregations.list,
        )
        self.delete = async_to_raw_response_wrapper(
            aggregations.delete,
        )


class AggregationsResourceWithStreamingResponse:
    def __init__(self, aggregations: AggregationsResource) -> None:
        self._aggregations = aggregations

        self.create = to_streamed_response_wrapper(
            aggregations.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            aggregations.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            aggregations.update,
        )
        self.list = to_streamed_response_wrapper(
            aggregations.list,
        )
        self.delete = to_streamed_response_wrapper(
            aggregations.delete,
        )


class AsyncAggregationsResourceWithStreamingResponse:
    def __init__(self, aggregations: AsyncAggregationsResource) -> None:
        self._aggregations = aggregations

        self.create = async_to_streamed_response_wrapper(
            aggregations.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            aggregations.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            aggregations.update,
        )
        self.list = async_to_streamed_response_wrapper(
            aggregations.list,
        )
        self.delete = async_to_streamed_response_wrapper(
            aggregations.delete,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Literal

import httpx

from ..._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from ..._utils import (
    maybe_transform,
    async_maybe_transform,
)
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...pagination import SyncCursor, AsyncCursor
from ...types.bills import debit_line_item_list_params, debit_line_item_create_params, debit_line_item_update_params
from ..._base_client import AsyncPaginator, make_request_options
from ...types.bills.debit_line_item_response import DebitLineItemResponse

__all__ = ["DebitLineItemsResource", "AsyncDebitLineItemsResource"]


class DebitLineItemsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> DebitLineItemsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#accessing-raw-response-data-eg-headers
        """
        return DebitLineItemsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> DebitLineItemsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#with_streaming_response
        """
        return DebitLineItemsResourceWithStreamingResponse(self)

    def create(
        self,
        bill_id: str,
        *,
        org_id: str | None = None,
        amount: float,
        description: str,
        product_id: str,
        referenced_bill_id: str,
        referenced_line_item_id: str,
        service_period_end_date: Union[str, datetime],
        service_period_start_date: Union[str, datetime],
        debit_reason_id: str | NotGiven = NOT_GIVEN,
        line_item_type: Literal[
            "STANDING_CHARGE",
            "USAGE",
            "COUNTER_RUNNING_TOTAL_CHARGE",
            "COUNTER_ADJUSTMENT_DEBIT",
            "COUNTER_ADJUSTMENT_CREDIT",
            "USAGE_CREDIT",
            "MINIMUM_SPEND",
            "MINIMUM_SPEND_REFUND",
            "CREDIT_DEDUCTION",
            "MANUAL_ADJUSTMENT",
            "CREDIT_MEMO",
            "DEBIT_MEMO",
            "COMMITMENT_CONSUMED",
            "COMMITMENT_FEE",
            "OVERAGE_SURCHARGE",
            "OVERAGE_USAGE",
            "BALANCE_CONSUMED",
            "BALANCE_FEE",
        ]
        | NotGiven = NOT_GIVEN,
        reason_id: str | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Create a new Debit line item for the given bill.

        When creating Debit line items for Bills, use the Debit Reasons created for your
        Organization. See [DebitReason](https://www.m3ter.com/docs/api#tag/DebitReason).

        Args:
          amount: The amount for the line item.

          description: The description for the line item.

          product_id: The UUID of the Product.

          referenced_bill_id: The UUID of the bill for the line item.

          referenced_line_item_id: The UUID of the line item.

          service_period_end_date: The service period end date in ISO-8601 format._(exclusive of the ending date)_.

          service_period_start_date: The service period start date in ISO-8601 format. _(inclusive of the starting
              date)_.

          debit_reason_id: The ID of the Debit Reason given for this debit line item.

          line_item_type

          reason_id: The UUID of the line item reason.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        return self._post(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems",
            body=maybe_transform(
                {
                    "amount": amount,
                    "description": description,
                    "product_id": product_id,
                    "referenced_bill_id": referenced_bill_id,
                    "referenced_line_item_id": referenced_line_item_id,
                    "service_period_end_date": service_period_end_date,
                    "service_period_start_date": service_period_start_date,
                    "debit_reason_id": debit_reason_id,
                    "line_item_type": line_item_type,
                    "reason_id": reason_id,
                    "version": version,
                },
                debit_line_item_create_params.DebitLineItemCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )

    def retrieve(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Retrieve the Debit line item with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._get(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )

    def update(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_id: str,
        amount: float,
        description: str,
        product_id: str,
        referenced_bill_id: str,
        referenced_line_item_id: str,
        service_period_end_date: Union[str, datetime],
        service_period_start_date: Union[str, datetime],
        debit_reason_id: str | NotGiven = NOT_GIVEN,
        line_item_type: Literal[
            "STANDING_CHARGE",
            "USAGE",
            "COUNTER_RUNNING_TOTAL_CHARGE",
            "COUNTER_ADJUSTMENT_DEBIT",
            "COUNTER_ADJUSTMENT_CREDIT",
            "USAGE_CREDIT",
            "MINIMUM_SPEND",
            "MINIMUM_SPEND_REFUND",
            "CREDIT_DEDUCTION",
            "MANUAL_ADJUSTMENT",
            "CREDIT_MEMO",
            "DEBIT_MEMO",
            "COMMITMENT_CONSUMED",
            "COMMITMENT_FEE",
            "OVERAGE_SURCHARGE",
            "OVERAGE_USAGE",
            "BALANCE_CONSUMED",
            "BALANCE_FEE",
        ]
        | NotGiven = NOT_GIVEN,
        reason_id: str | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Update the Debit line item with the given UUID.

        Args:
          amount: The amount for the line item.

          description: The description for the line item.

          product_id: The UUID of the Product.

          referenced_bill_id: The UUID of the bill for the line item.

          referenced_line_item_id: The UUID of the line item.

          service_period_end_date: The service period end date in ISO-8601 format._(exclusive of the ending date)_.

          service_period_start_date: The service period start date in ISO-8601 format. _(inclusive of the starting
              date)_.

          debit_reason_id: The ID of the Debit Reason given for this debit line item.

          line_item_type

          reason_id: The UUID of the line item reason.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._put(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems/{id}",
            body=maybe_transform(
                {
                    "amount": amount,
                    "description": description,
                    "product_id": product_id,
                    "referenced_bill_id": referenced_bill_id,
                    "referenced_line_item_id": referenced_line_item_id,
                    "service_period_end_date": service_period_end_date,
                    "service_period_start_date": service_period_start_date,
                    "debit_reason_id": debit_reason_id,
                    "line_item_type": line_item_type,
                    "reason_id": reason_id,
                    "version": version,
                },
                debit_line_item_update_params.DebitLineItemUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )

    def list(
        self,
        bill_id: str,
        *,
        org_id: str | None = None,
        next_token: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncCursor[DebitLineItemResponse]:
        """
        List the Debit line items for the given bill.

        Args:
          next_token: `nextToken` for multi page retrievals.

          page_size: Number of line items to retrieve per page.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        return self._get_api_list(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems",
            page=SyncCursor[DebitLineItemResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "next_token": next_token,
                        "page_size": page_size,
                    },
                    debit_line_item_list_params.DebitLineItemListParams,
                ),
            ),
            model=DebitLineItemResponse,
        )

    def delete(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Delete the Debit line item with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._delete(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )


class AsyncDebitLineItemsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncDebitLineItemsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#accessing-raw-response-data-eg-headers
        """
        return AsyncDebitLineItemsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncDebitLineItemsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#with_streaming_response
        """
        return AsyncDebitLineItemsResourceWithStreamingResponse(self)

    async def create(
        self,
        bill_id: str,
        *,
        org_id: str | None = None,
        amount: float,
        description: str,
        product_id: str,
        referenced_bill_id: str,
        referenced_line_item_id: str,
        service_period_end_date: Union[str, datetime],
        service_period_start_date: Union[str, datetime],
        debit_reason_id: str | NotGiven = NOT_GIVEN,
        line_item_type: Literal[
            "STANDING_CHARGE",
            "USAGE",
            "COUNTER_RUNNING_TOTAL_CHARGE",
            "COUNTER_ADJUSTMENT_DEBIT",
            "COUNTER_ADJUSTMENT_CREDIT",
            "USAGE_CREDIT",
            "MINIMUM_SPEND",
            "MINIMUM_SPEND_REFUND",
            "CREDIT_DEDUCTION",
            "MANUAL_ADJUSTMENT",
            "CREDIT_MEMO",
            "DEBIT_MEMO",
            "COMMITMENT_CONSUMED",
            "COMMITMENT_FEE",
            "OVERAGE_SURCHARGE",
            "OVERAGE_USAGE",
            "BALANCE_CONSUMED",
            "BALANCE_FEE",
        ]
        | NotGiven = NOT_GIVEN,
        reason_id: str | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Create a new Debit line item for the given bill.

        When creating Debit line items for Bills, use the Debit Reasons created for your
        Organization. See [DebitReason](https://www.m3ter.com/docs/api#tag/DebitReason).

        Args:
          amount: The amount for the line item.

          description: The description for the line item.

          product_id: The UUID of the Product.

          referenced_bill_id: The UUID of the bill for the line item.

          referenced_line_item_id: The UUID of the line item.

          service_period_end_date: The service period end date in ISO-8601 format._(exclusive of the ending date)_.

          service_period_start_date: The service period start date in ISO-8601 format. _(inclusive of the starting
              date)_.

          debit_reason_id: The ID of the Debit Reason given for this debit line item.

          line_item_type

          reason_id: The UUID of the line item reason.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        return await self._post(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems",
            body=await async_maybe_transform(
                {
                    "amount": amount,
                    "description": description,
                    "product_id": product_id,
                    "referenced_bill_id": referenced_bill_id,
                    "referenced_line_item_id": referenced_line_item_id,
                    "service_period_end_date": service_period_end_date,
                    "service_period_start_date": service_period_start_date,
                    "debit_reason_id": debit_reason_id,
                    "line_item_type": line_item_type,
                    "reason_id": reason_id,
                    "version": version,
                },
                debit_line_item_create_params.DebitLineItemCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )

    async def retrieve(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Retrieve the Debit line item with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._get(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )

    async def update(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_id: str,
        amount: float,
        description: str,
        product_id: str,
        referenced_bill_id: str,
        referenced_line_item_id: str,
        service_period_end_date: Union[str, datetime],
        service_period_start_date: Union[str, datetime],
        debit_reason_id: str | NotGiven = NOT_GIVEN,
        line_item_type: Literal[
            "STANDING_CHARGE",
            "USAGE",
            "COUNTER_RUNNING_TOTAL_CHARGE",
            "COUNTER_ADJUSTMENT_DEBIT",
            "COUNTER_ADJUSTMENT_CREDIT",
            "USAGE_CREDIT",
            "MINIMUM_SPEND",
            "MINIMUM_SPEND_REFUND",
            "CREDIT_DEDUCTION",
            "MANUAL_ADJUSTMENT",
            "CREDIT_MEMO",
            "DEBIT_MEMO",
            "COMMITMENT_CONSUMED",
            "COMMITMENT_FEE",
            "OVERAGE_SURCHARGE",
            "OVERAGE_USAGE",
            "BALANCE_CONSUMED",
            "BALANCE_FEE",
        ]
        | NotGiven = NOT_GIVEN,
        reason_id: str | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Update the Debit line item with the given UUID.

        Args:
          amount: The amount for the line item.

          description: The description for the line item.

          product_id: The UUID of the Product.

          referenced_bill_id: The UUID of the bill for the line item.

          referenced_line_item_id: The UUID of the line item.

          service_period_end_date: The service period end date in ISO-8601 format._(exclusive of the ending date)_.

          service_period_start_date: The service period start date in ISO-8601 format. _(inclusive of the starting
              date)_.

          debit_reason_id: The ID of the Debit Reason given for this debit line item.

          line_item_type

          reason_id: The UUID of the line item reason.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._put(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems/{id}",
            body=await async_maybe_transform(
                {
                    "amount": amount,
                    "description": description,
                    "product_id": product_id,
                    "referenced_bill_id": referenced_bill_id,
                    "referenced_line_item_id": referenced_line_item_id,
                    "service_period_end_date": service_period_end_date,
                    "service_period_start_date": service_period_start_date,
                    "debit_reason_id": debit_reason_id,
                    "line_item_type": line_item_type,
                    "reason_id": reason_id,
                    "version": version,
                },
                debit_line_item_update_params.DebitLineItemUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )

    def list(
        self,
        bill_id: str,
        *,
        org_id: str | None = None,
        next_token: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[DebitLineItemResponse, AsyncCursor[DebitLineItemResponse]]:
        """
        List the Debit line items for the given bill.

        Args:
          next_token: `nextToken` for multi page retrievals.

          page_size: Number of line items to retrieve per page.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        return self._get_api_list(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems",
            page=AsyncCursor[DebitLineItemResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "next_token": next_token,
                        "page_size": page_size,
                    },
                    debit_line_item_list_params.DebitLineItemListParams,
                ),
            ),
            model=DebitLineItemResponse,
        )

    async def delete(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> DebitLineItemResponse:
        """
        Delete the Debit line item with the given UUID.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not bill_id:
            raise ValueError(f"Expected a non-empty value for `bill_id` but received {bill_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._delete(
            f"/organizations/{org_id}/bills/{bill_id}/debitlineitems/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DebitLineItemResponse,
        )


class DebitLineItemsResourceWithRawResponse:
    def __init__(self, debit_line_items: DebitLineItemsResource) -> None:
        self._debit_line_items = debit_line_items

        self.create = to_raw_response_wrapper(
            debit_line_items.create,
        )
        self.retrieve = to_raw_response_wrapper(
            debit_line_items.retrieve,
        )
        self.update = to_raw_response_wrapper(
            debit_line_items.update,
        )
        self.list = to_raw_response_wrapper(
            debit_line_items.list,
        )
        self.delete = to_raw_response_wrapper(
            debit_line_items.delete,
        )


class AsyncDebitLineItemsResourceWithRawResponse:
    def __init__(self, debit_line_items: AsyncDebitLineItemsResource) -> None:
        self._debit_line_items = debit_line_items

        self.create = async_to_raw_response_wrapper(
            debit_line_items.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            debit_line_items.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            debit_line_items.update,
        )
        self.list = async_to_raw_response_wrapper(
            debit_line_items.list,
        )
        self.delete = async_to_raw_response_wrapper(
            debit_line_items.delete,
        )


class DebitLineItemsResourceWithStreamingResponse:
    def __init__(self, debit_line_items: DebitLineItemsResource) -> None:
        self._debit_line_items = debit_line_items

        self.create = to_streamed_response_wrapper(
            debit_line_items.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            debit_line_items.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            debit_line_items.update,
        )
        self.list = to_streamed_response_wrapper(
            debit_line_items.list,
        )
        self.delete = to_streamed_response_wrapper(
            debit_line_items.delete,
        )


class AsyncDebitLineItemsResourceWithStreamingResponse:
    def __init__(self, debit_line_items: AsyncDebitLineItemsResource) -> None:
        self._debit_line_items = debit_line_items

        self.create = async_to_streamed_response_wrapper(
            debit_line_items.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            debit_line_items.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            debit_line_items.update,
        )
        self.list = async_to_streamed_response_wrapper(
            debit_line_items.list,
        )
        self.delete = async_to_streamed_response_wrapper(
            debit_line_items.delete,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union
from typing_extensions import Literal

import httpx

from ..types import plan_template_list_params, plan_template_create_params, plan_template_update_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import (
    maybe_transform,
    async_maybe_transform,
)
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ..pagination import SyncCursor, AsyncCursor
from .._base_client import AsyncPaginator, make_request_options
from ..types.plan_template_response import PlanTemplateResponse

__all__ = ["PlanTemplatesResource", "AsyncPlanTemplatesResource"]


class PlanTemplatesResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> PlanTemplatesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#accessing-raw-response-data-eg-headers
        """
        return PlanTemplatesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> PlanTemplatesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#with_streaming_response
        """
        return PlanTemplatesResourceWithStreamingResponse(self)

    def create(
        self,
        *,
        org_id: str | None = None,
        bill_frequency: Literal["DAILY", "WEEKLY", "MONTHLY", "ANNUALLY", "AD_HOC", "MIXED"],
        currency: str,
        name: str,
        product_id: str,
        standing_charge: float,
        bill_frequency_interval: int | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        minimum_spend: float | NotGiven = NOT_GIVEN,
        minimum_spend_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        minimum_spend_description: str | NotGiven = NOT_GIVEN,
        ordinal: int | NotGiven = NOT_GIVEN,
        standing_charge_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        standing_charge_description: str | NotGiven = NOT_GIVEN,
        standing_charge_interval: int | NotGiven = NOT_GIVEN,
        standing_charge_offset: int | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Create a new PlanTemplate.

        This endpoint creates a new PlanTemplate within a specific Organization,
        identified by its unique UUID. The request body should contain the necessary
        information for the new PlanTemplate.

        Args:
          bill_frequency: Determines the frequency at which bills are generated.

              - **Daily**. Starting at midnight each day, covering the twenty-four hour period
                following.

              - **Weekly**. Starting at midnight on a Monday, covering the seven-day period
                following.

              - **Monthly**. Starting at midnight on the first day of each month, covering the
                entire calendar month following.

              - **Annually**. Starting at midnight on first day of each year covering the
                entire calendar year following.

          currency: The ISO currency code for the currency used to charge end users - for example
              USD, GBP, EUR. This defines the _pricing currency_ and is inherited by any Plans
              based on the Plan Template.

              **Notes:**

              - You can define a currency at Organization-level or Account-level to be used as
                the _billing currency_. This can be a different currency to that used for the
                Plan as the _pricing currency_.
              - If the billing currency for an Account is different to the pricing currency
                used by a Plan attached to the Account, you must ensure a _currency conversion
                rate_ is defined for your Organization to convert the pricing currency into
                the billing currency at billing, otherwise Bills will fail for the Account.
              - To define any required currency conversion rates, use the
                `currencyConversions` request body parameter for the
                [Update OrganizationConfig](https://www.m3ter.com/docs/api#tag/OrganizationConfig/operation/UpdateOrganizationConfig)
                call.

          name: Descriptive name for the PlanTemplate.

          product_id: The unique identifier (UUID) of the Product associated with this PlanTemplate.

          standing_charge: The fixed charge _(standing charge)_ applied to customer bills. This charge is
              prorated and must be a non-negative number.

          bill_frequency_interval: How often bills are issued. For example, if `billFrequency` is Monthly and
              `billFrequencyInterval` is 3, bills are issued every three months.

          code: A unique, short code reference for the PlanTemplate. This code should not
              contain control characters or spaces.

          custom_fields: User defined fields enabling you to attach custom data. The value for a custom
              field can be either a string or a number.

              If `customFields` can also be defined for this entity at the Organizational
              level, `customField` values defined at individual level override values of
              `customFields` with the same name defined at Organization level.

              See
              [Working with Custom Fields](https://www.m3ter.com/docs/guides/creating-and-managing-products/working-with-custom-fields)
              in the m3ter documentation for more information.

          minimum_spend: The Product minimum spend amount per billing cycle for end customer Accounts on
              a pricing Plan based on the PlanTemplate. This must be a non-negative number.

          minimum_spend_bill_in_advance: A boolean that determines when the minimum spend is billed.

              - TRUE - minimum spend is billed at the start of each billing period.
              - FALSE - minimum spend is billed at the end of each billing period.

              Overrides the setting at Organizational level for minimum spend billing in
              arrears/in advance.

          minimum_spend_description: Minimum spend description _(displayed on the bill line item)_.

          ordinal: The ranking of the PlanTemplate among your pricing plans. Lower numbers
              represent more basic plans, while higher numbers represent premium plans. This
              must be a non-negative integer.

              **NOTE: DEPRECATED** - do not use.

          standing_charge_bill_in_advance: A boolean that determines when the standing charge is billed.

              - TRUE - standing charge is billed at the start of each billing period.
              - FALSE - standing charge is billed at the end of each billing period.

              Overrides the setting at Organizational level for standing charge billing in
              arrears/in advance.

          standing_charge_description: Standing charge description _(displayed on the bill line item)_.

          standing_charge_interval: How often the standing charge is applied. For example, if the bill is issued
              every three months and `standingChargeInterval` is 2, then the standing charge
              is applied every six months.

          standing_charge_offset: Defines an offset for when the standing charge is first applied. For example, if
              the bill is issued every three months and the `standingChargeOfset` is 0, then
              the charge is applied to the first bill _(at three months)_; if 1, it would be
              applied to the second bill _(at six months)_, and so on.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return self._post(
            f"/organizations/{org_id}/plantemplates",
            body=maybe_transform(
                {
                    "bill_frequency": bill_frequency,
                    "currency": currency,
                    "name": name,
                    "product_id": product_id,
                    "standing_charge": standing_charge,
                    "bill_frequency_interval": bill_frequency_interval,
                    "code": code,
                    "custom_fields": custom_fields,
                    "minimum_spend": minimum_spend,
                    "minimum_spend_bill_in_advance": minimum_spend_bill_in_advance,
                    "minimum_spend_description": minimum_spend_description,
                    "ordinal": ordinal,
                    "standing_charge_bill_in_advance": standing_charge_bill_in_advance,
                    "standing_charge_description": standing_charge_description,
                    "standing_charge_interval": standing_charge_interval,
                    "standing_charge_offset": standing_charge_offset,
                    "version": version,
                },
                plan_template_create_params.PlanTemplateCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )

    def retrieve(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Retrieve a specific PlanTemplate.

        This endpoint allows you to retrieve a specific PlanTemplate within a specific
        Organization, both identified by their unique identifiers (UUIDs).

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._get(
            f"/organizations/{org_id}/plantemplates/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )

    def update(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_frequency: Literal["DAILY", "WEEKLY", "MONTHLY", "ANNUALLY", "AD_HOC", "MIXED"],
        currency: str,
        name: str,
        product_id: str,
        standing_charge: float,
        bill_frequency_interval: int | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        minimum_spend: float | NotGiven = NOT_GIVEN,
        minimum_spend_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        minimum_spend_description: str | NotGiven = NOT_GIVEN,
        ordinal: int | NotGiven = NOT_GIVEN,
        standing_charge_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        standing_charge_description: str | NotGiven = NOT_GIVEN,
        standing_charge_interval: int | NotGiven = NOT_GIVEN,
        standing_charge_offset: int | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Update a specific PlanTemplate.

        This endpoint enables you to update a specific PlanTemplate within a specific
        Organization, both identified by their unique identifiers (UUIDs). The request
        body should contain the updated information for the PlanTemplate.

        **Note:** If you have created Custom Fields for a Plan Template, when you use
        this endpoint to update the Plan Template use the `customFields` parameter to
        preserve those Custom Fields. If you omit them from the update request, they
        will be lost.

        Args:
          bill_frequency: Determines the frequency at which bills are generated.

              - **Daily**. Starting at midnight each day, covering the twenty-four hour period
                following.

              - **Weekly**. Starting at midnight on a Monday, covering the seven-day period
                following.

              - **Monthly**. Starting at midnight on the first day of each month, covering the
                entire calendar month following.

              - **Annually**. Starting at midnight on first day of each year covering the
                entire calendar year following.

          currency: The ISO currency code for the currency used to charge end users - for example
              USD, GBP, EUR. This defines the _pricing currency_ and is inherited by any Plans
              based on the Plan Template.

              **Notes:**

              - You can define a currency at Organization-level or Account-level to be used as
                the _billing currency_. This can be a different currency to that used for the
                Plan as the _pricing currency_.
              - If the billing currency for an Account is different to the pricing currency
                used by a Plan attached to the Account, you must ensure a _currency conversion
                rate_ is defined for your Organization to convert the pricing currency into
                the billing currency at billing, otherwise Bills will fail for the Account.
              - To define any required currency conversion rates, use the
                `currencyConversions` request body parameter for the
                [Update OrganizationConfig](https://www.m3ter.com/docs/api#tag/OrganizationConfig/operation/UpdateOrganizationConfig)
                call.

          name: Descriptive name for the PlanTemplate.

          product_id: The unique identifier (UUID) of the Product associated with this PlanTemplate.

          standing_charge: The fixed charge _(standing charge)_ applied to customer bills. This charge is
              prorated and must be a non-negative number.

          bill_frequency_interval: How often bills are issued. For example, if `billFrequency` is Monthly and
              `billFrequencyInterval` is 3, bills are issued every three months.

          code: A unique, short code reference for the PlanTemplate. This code should not
              contain control characters or spaces.

          custom_fields: User defined fields enabling you to attach custom data. The value for a custom
              field can be either a string or a number.

              If `customFields` can also be defined for this entity at the Organizational
              level, `customField` values defined at individual level override values of
              `customFields` with the same name defined at Organization level.

              See
              [Working with Custom Fields](https://www.m3ter.com/docs/guides/creating-and-managing-products/working-with-custom-fields)
              in the m3ter documentation for more information.

          minimum_spend: The Product minimum spend amount per billing cycle for end customer Accounts on
              a pricing Plan based on the PlanTemplate. This must be a non-negative number.

          minimum_spend_bill_in_advance: A boolean that determines when the minimum spend is billed.

              - TRUE - minimum spend is billed at the start of each billing period.
              - FALSE - minimum spend is billed at the end of each billing period.

              Overrides the setting at Organizational level for minimum spend billing in
              arrears/in advance.

          minimum_spend_description: Minimum spend description _(displayed on the bill line item)_.

          ordinal: The ranking of the PlanTemplate among your pricing plans. Lower numbers
              represent more basic plans, while higher numbers represent premium plans. This
              must be a non-negative integer.

              **NOTE: DEPRECATED** - do not use.

          standing_charge_bill_in_advance: A boolean that determines when the standing charge is billed.

              - TRUE - standing charge is billed at the start of each billing period.
              - FALSE - standing charge is billed at the end of each billing period.

              Overrides the setting at Organizational level for standing charge billing in
              arrears/in advance.

          standing_charge_description: Standing charge description _(displayed on the bill line item)_.

          standing_charge_interval: How often the standing charge is applied. For example, if the bill is issued
              every three months and `standingChargeInterval` is 2, then the standing charge
              is applied every six months.

          standing_charge_offset: Defines an offset for when the standing charge is first applied. For example, if
              the bill is issued every three months and the `standingChargeOfset` is 0, then
              the charge is applied to the first bill _(at three months)_; if 1, it would be
              applied to the second bill _(at six months)_, and so on.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._put(
            f"/organizations/{org_id}/plantemplates/{id}",
            body=maybe_transform(
                {
                    "bill_frequency": bill_frequency,
                    "currency": currency,
                    "name": name,
                    "product_id": product_id,
                    "standing_charge": standing_charge,
                    "bill_frequency_interval": bill_frequency_interval,
                    "code": code,
                    "custom_fields": custom_fields,
                    "minimum_spend": minimum_spend,
                    "minimum_spend_bill_in_advance": minimum_spend_bill_in_advance,
                    "minimum_spend_description": minimum_spend_description,
                    "ordinal": ordinal,
                    "standing_charge_bill_in_advance": standing_charge_bill_in_advance,
                    "standing_charge_description": standing_charge_description,
                    "standing_charge_interval": standing_charge_interval,
                    "standing_charge_offset": standing_charge_offset,
                    "version": version,
                },
                plan_template_update_params.PlanTemplateUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )

    def list(
        self,
        *,
        org_id: str | None = None,
        ids: List[str] | NotGiven = NOT_GIVEN,
        next_token: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        product_id: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncCursor[PlanTemplateResponse]:
        """
        Retrieve a list of PlanTemplates.

        This endpoint enables you to retrieve a paginated list of PlanTemplates
        belonging to a specific Organization, identified by its UUID. You can filter the
        list by PlanTemplate IDs or Product IDs for more focused retrieval.

        Args:
          ids: List of specific PlanTemplate UUIDs to retrieve.

          next_token: The `nextToken` for multi-page retrievals. It is used to fetch the next page of
              PlanTemplates in a paginated list.

          page_size: Specifies the maximum number of PlanTemplates to retrieve per page.

          product_id: The unique identifiers (UUIDs) of the Products to retrieve associated
              PlanTemplates.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return self._get_api_list(
            f"/organizations/{org_id}/plantemplates",
            page=SyncCursor[PlanTemplateResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "ids": ids,
                        "next_token": next_token,
                        "page_size": page_size,
                        "product_id": product_id,
                    },
                    plan_template_list_params.PlanTemplateListParams,
                ),
            ),
            model=PlanTemplateResponse,
        )

    def delete(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Delete a specific PlanTemplate.

        This endpoint enables you to delete a specific PlanTemplate within a specific
        Organization, both identified by their unique identifiers (UUIDs).

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._delete(
            f"/organizations/{org_id}/plantemplates/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )


class AsyncPlanTemplatesResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncPlanTemplatesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#accessing-raw-response-data-eg-headers
        """
        return AsyncPlanTemplatesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncPlanTemplatesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/m3ter-com/m3ter-sdk-python#with_streaming_response
        """
        return AsyncPlanTemplatesResourceWithStreamingResponse(self)

    async def create(
        self,
        *,
        org_id: str | None = None,
        bill_frequency: Literal["DAILY", "WEEKLY", "MONTHLY", "ANNUALLY", "AD_HOC", "MIXED"],
        currency: str,
        name: str,
        product_id: str,
        standing_charge: float,
        bill_frequency_interval: int | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        minimum_spend: float | NotGiven = NOT_GIVEN,
        minimum_spend_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        minimum_spend_description: str | NotGiven = NOT_GIVEN,
        ordinal: int | NotGiven = NOT_GIVEN,
        standing_charge_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        standing_charge_description: str | NotGiven = NOT_GIVEN,
        standing_charge_interval: int | NotGiven = NOT_GIVEN,
        standing_charge_offset: int | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Create a new PlanTemplate.

        This endpoint creates a new PlanTemplate within a specific Organization,
        identified by its unique UUID. The request body should contain the necessary
        information for the new PlanTemplate.

        Args:
          bill_frequency: Determines the frequency at which bills are generated.

              - **Daily**. Starting at midnight each day, covering the twenty-four hour period
                following.

              - **Weekly**. Starting at midnight on a Monday, covering the seven-day period
                following.

              - **Monthly**. Starting at midnight on the first day of each month, covering the
                entire calendar month following.

              - **Annually**. Starting at midnight on first day of each year covering the
                entire calendar year following.

          currency: The ISO currency code for the currency used to charge end users - for example
              USD, GBP, EUR. This defines the _pricing currency_ and is inherited by any Plans
              based on the Plan Template.

              **Notes:**

              - You can define a currency at Organization-level or Account-level to be used as
                the _billing currency_. This can be a different currency to that used for the
                Plan as the _pricing currency_.
              - If the billing currency for an Account is different to the pricing currency
                used by a Plan attached to the Account, you must ensure a _currency conversion
                rate_ is defined for your Organization to convert the pricing currency into
                the billing currency at billing, otherwise Bills will fail for the Account.
              - To define any required currency conversion rates, use the
                `currencyConversions` request body parameter for the
                [Update OrganizationConfig](https://www.m3ter.com/docs/api#tag/OrganizationConfig/operation/UpdateOrganizationConfig)
                call.

          name: Descriptive name for the PlanTemplate.

          product_id: The unique identifier (UUID) of the Product associated with this PlanTemplate.

          standing_charge: The fixed charge _(standing charge)_ applied to customer bills. This charge is
              prorated and must be a non-negative number.

          bill_frequency_interval: How often bills are issued. For example, if `billFrequency` is Monthly and
              `billFrequencyInterval` is 3, bills are issued every three months.

          code: A unique, short code reference for the PlanTemplate. This code should not
              contain control characters or spaces.

          custom_fields: User defined fields enabling you to attach custom data. The value for a custom
              field can be either a string or a number.

              If `customFields` can also be defined for this entity at the Organizational
              level, `customField` values defined at individual level override values of
              `customFields` with the same name defined at Organization level.

              See
              [Working with Custom Fields](https://www.m3ter.com/docs/guides/creating-and-managing-products/working-with-custom-fields)
              in the m3ter documentation for more information.

          minimum_spend: The Product minimum spend amount per billing cycle for end customer Accounts on
              a pricing Plan based on the PlanTemplate. This must be a non-negative number.

          minimum_spend_bill_in_advance: A boolean that determines when the minimum spend is billed.

              - TRUE - minimum spend is billed at the start of each billing period.
              - FALSE - minimum spend is billed at the end of each billing period.

              Overrides the setting at Organizational level for minimum spend billing in
              arrears/in advance.

          minimum_spend_description: Minimum spend description _(displayed on the bill line item)_.

          ordinal: The ranking of the PlanTemplate among your pricing plans. Lower numbers
              represent more basic plans, while higher numbers represent premium plans. This
              must be a non-negative integer.

              **NOTE: DEPRECATED** - do not use.

          standing_charge_bill_in_advance: A boolean that determines when the standing charge is billed.

              - TRUE - standing charge is billed at the start of each billing period.
              - FALSE - standing charge is billed at the end of each billing period.

              Overrides the setting at Organizational level for standing charge billing in
              arrears/in advance.

          standing_charge_description: Standing charge description _(displayed on the bill line item)_.

          standing_charge_interval: How often the standing charge is applied. For example, if the bill is issued
              every three months and `standingChargeInterval` is 2, then the standing charge
              is applied every six months.

          standing_charge_offset: Defines an offset for when the standing charge is first applied. For example, if
              the bill is issued every three months and the `standingChargeOfset` is 0, then
              the charge is applied to the first bill _(at three months)_; if 1, it would be
              applied to the second bill _(at six months)_, and so on.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return await self._post(
            f"/organizations/{org_id}/plantemplates",
            body=await async_maybe_transform(
                {
                    "bill_frequency": bill_frequency,
                    "currency": currency,
                    "name": name,
                    "product_id": product_id,
                    "standing_charge": standing_charge,
                    "bill_frequency_interval": bill_frequency_interval,
                    "code": code,
                    "custom_fields": custom_fields,
                    "minimum_spend": minimum_spend,
                    "minimum_spend_bill_in_advance": minimum_spend_bill_in_advance,
                    "minimum_spend_description": minimum_spend_description,
                    "ordinal": ordinal,
                    "standing_charge_bill_in_advance": standing_charge_bill_in_advance,
                    "standing_charge_description": standing_charge_description,
                    "standing_charge_interval": standing_charge_interval,
                    "standing_charge_offset": standing_charge_offset,
                    "version": version,
                },
                plan_template_create_params.PlanTemplateCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )

    async def retrieve(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Retrieve a specific PlanTemplate.

        This endpoint allows you to retrieve a specific PlanTemplate within a specific
        Organization, both identified by their unique identifiers (UUIDs).

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._get(
            f"/organizations/{org_id}/plantemplates/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )

    async def update(
        self,
        id: str,
        *,
        org_id: str | None = None,
        bill_frequency: Literal["DAILY", "WEEKLY", "MONTHLY", "ANNUALLY", "AD_HOC", "MIXED"],
        currency: str,
        name: str,
        product_id: str,
        standing_charge: float,
        bill_frequency_interval: int | NotGiven = NOT_GIVEN,
        code: str | NotGiven = NOT_GIVEN,
        custom_fields: Dict[str, Union[str, float]] | NotGiven = NOT_GIVEN,
        minimum_spend: float | NotGiven = NOT_GIVEN,
        minimum_spend_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        minimum_spend_description: str | NotGiven = NOT_GIVEN,
        ordinal: int | NotGiven = NOT_GIVEN,
        standing_charge_bill_in_advance: bool | NotGiven = NOT_GIVEN,
        standing_charge_description: str | NotGiven = NOT_GIVEN,
        standing_charge_interval: int | NotGiven = NOT_GIVEN,
        standing_charge_offset: int | NotGiven = NOT_GIVEN,
        version: int | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Update a specific PlanTemplate.

        This endpoint enables you to update a specific PlanTemplate within a specific
        Organization, both identified by their unique identifiers (UUIDs). The request
        body should contain the updated information for the PlanTemplate.

        **Note:** If you have created Custom Fields for a Plan Template, when you use
        this endpoint to update the Plan Template use the `customFields` parameter to
        preserve those Custom Fields. If you omit them from the update request, they
        will be lost.

        Args:
          bill_frequency: Determines the frequency at which bills are generated.

              - **Daily**. Starting at midnight each day, covering the twenty-four hour period
                following.

              - **Weekly**. Starting at midnight on a Monday, covering the seven-day period
                following.

              - **Monthly**. Starting at midnight on the first day of each month, covering the
                entire calendar month following.

              - **Annually**. Starting at midnight on first day of each year covering the
                entire calendar year following.

          currency: The ISO currency code for the currency used to charge end users - for example
              USD, GBP, EUR. This defines the _pricing currency_ and is inherited by any Plans
              based on the Plan Template.

              **Notes:**

              - You can define a currency at Organization-level or Account-level to be used as
                the _billing currency_. This can be a different currency to that used for the
                Plan as the _pricing currency_.
              - If the billing currency for an Account is different to the pricing currency
                used by a Plan attached to the Account, you must ensure a _currency conversion
                rate_ is defined for your Organization to convert the pricing currency into
                the billing currency at billing, otherwise Bills will fail for the Account.
              - To define any required currency conversion rates, use the
                `currencyConversions` request body parameter for the
                [Update OrganizationConfig](https://www.m3ter.com/docs/api#tag/OrganizationConfig/operation/UpdateOrganizationConfig)
                call.

          name: Descriptive name for the PlanTemplate.

          product_id: The unique identifier (UUID) of the Product associated with this PlanTemplate.

          standing_charge: The fixed charge _(standing charge)_ applied to customer bills. This charge is
              prorated and must be a non-negative number.

          bill_frequency_interval: How often bills are issued. For example, if `billFrequency` is Monthly and
              `billFrequencyInterval` is 3, bills are issued every three months.

          code: A unique, short code reference for the PlanTemplate. This code should not
              contain control characters or spaces.

          custom_fields: User defined fields enabling you to attach custom data. The value for a custom
              field can be either a string or a number.

              If `customFields` can also be defined for this entity at the Organizational
              level, `customField` values defined at individual level override values of
              `customFields` with the same name defined at Organization level.

              See
              [Working with Custom Fields](https://www.m3ter.com/docs/guides/creating-and-managing-products/working-with-custom-fields)
              in the m3ter documentation for more information.

          minimum_spend: The Product minimum spend amount per billing cycle for end customer Accounts on
              a pricing Plan based on the PlanTemplate. This must be a non-negative number.

          minimum_spend_bill_in_advance: A boolean that determines when the minimum spend is billed.

              - TRUE - minimum spend is billed at the start of each billing period.
              - FALSE - minimum spend is billed at the end of each billing period.

              Overrides the setting at Organizational level for minimum spend billing in
              arrears/in advance.

          minimum_spend_description: Minimum spend description _(displayed on the bill line item)_.

          ordinal: The ranking of the PlanTemplate among your pricing plans. Lower numbers
              represent more basic plans, while higher numbers represent premium plans. This
              must be a non-negative integer.

              **NOTE: DEPRECATED** - do not use.

          standing_charge_bill_in_advance: A boolean that determines when the standing charge is billed.

              - TRUE - standing charge is billed at the start of each billing period.
              - FALSE - standing charge is billed at the end of each billing period.

              Overrides the setting at Organizational level for standing charge billing in
              arrears/in advance.

          standing_charge_description: Standing charge description _(displayed on the bill line item)_.

          standing_charge_interval: How often the standing charge is applied. For example, if the bill is issued
              every three months and `standingChargeInterval` is 2, then the standing charge
              is applied every six months.

          standing_charge_offset: Defines an offset for when the standing charge is first applied. For example, if
              the bill is issued every three months and the `standingChargeOfset` is 0, then
              the charge is applied to the first bill _(at three months)_; if 1, it would be
              applied to the second bill _(at six months)_, and so on.

          version:
              The version number of the entity:

              - **Create entity:** Not valid for initial insertion of new entity - _do not use
                for Create_. On initial Create, version is set at 1 and listed in the
                response.
              - **Update Entity:** On Update, version is required and must match the existing
                version because a check is performed to ensure sequential versioning is
                preserved. Version is incremented by 1 and listed in the response.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._put(
            f"/organizations/{org_id}/plantemplates/{id}",
            body=await async_maybe_transform(
                {
                    "bill_frequency": bill_frequency,
                    "currency": currency,
                    "name": name,
                    "product_id": product_id,
                    "standing_charge": standing_charge,
                    "bill_frequency_interval": bill_frequency_interval,
                    "code": code,
                    "custom_fields": custom_fields,
                    "minimum_spend": minimum_spend,
                    "minimum_spend_bill_in_advance": minimum_spend_bill_in_advance,
                    "minimum_spend_description": minimum_spend_description,
                    "ordinal": ordinal,
                    "standing_charge_bill_in_advance": standing_charge_bill_in_advance,
                    "standing_charge_description": standing_charge_description,
                    "standing_charge_interval": standing_charge_interval,
                    "standing_charge_offset": standing_charge_offset,
                    "version": version,
                },
                plan_template_update_params.PlanTemplateUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )

    def list(
        self,
        *,
        org_id: str | None = None,
        ids: List[str] | NotGiven = NOT_GIVEN,
        next_token: str | NotGiven = NOT_GIVEN,
        page_size: int | NotGiven = NOT_GIVEN,
        product_id: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[PlanTemplateResponse, AsyncCursor[PlanTemplateResponse]]:
        """
        Retrieve a list of PlanTemplates.

        This endpoint enables you to retrieve a paginated list of PlanTemplates
        belonging to a specific Organization, identified by its UUID. You can filter the
        list by PlanTemplate IDs or Product IDs for more focused retrieval.

        Args:
          ids: List of specific PlanTemplate UUIDs to retrieve.

          next_token: The `nextToken` for multi-page retrievals. It is used to fetch the next page of
              PlanTemplates in a paginated list.

          page_size: Specifies the maximum number of PlanTemplates to retrieve per page.

          product_id: The unique identifiers (UUIDs) of the Products to retrieve associated
              PlanTemplates.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        return self._get_api_list(
            f"/organizations/{org_id}/plantemplates",
            page=AsyncCursor[PlanTemplateResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "ids": ids,
                        "next_token": next_token,
                        "page_size": page_size,
                        "product_id": product_id,
                    },
                    plan_template_list_params.PlanTemplateListParams,
                ),
            ),
            model=PlanTemplateResponse,
        )

    async def delete(
        self,
        id: str,
        *,
        org_id: str | None = None,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> PlanTemplateResponse:
        """
        Delete a specific PlanTemplate.

        This endpoint enables you to delete a specific PlanTemplate within a specific
        Organization, both identified by their unique identifiers (UUIDs).

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if org_id is None:
            org_id = self._client._get_org_id_path_param()
        if not org_id:
            raise ValueError(f"Expected a non-empty value for `org_id` but received {org_id!r}")
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._delete(
            f"/organizations/{org_id}/plantemplates/{id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=PlanTemplateResponse,
        )


class PlanTemplatesResourceWithRawResponse:
    def __init__(self, plan_templates: PlanTemplatesResource) -> None:
        self._plan_templates = plan_templates

        self.create = to_raw_response_wrapper(
            plan_templates.create,
        )
        self.retrieve = to_raw_response_wrapper(
            plan_templates.retrieve,
        )
        self.update = to_raw_response_wrapper(
            plan_templates.update,
        )
        self.list = to_raw_response_wrapper(
            plan_templates.list,
        )
        self.delete = to_raw_response_wrapper(
            plan_templates.delete,
        )


class AsyncPlanTemplatesResourceWithRawResponse:
    def __init__(self, plan_templates: AsyncPlanTemplatesResource) -> None:
        self._plan_templates = plan_templates

        self.create = async_to_raw_response_wrapper(
            plan_templates.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            plan_templates.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            plan_templates.update,
        )
        self.list = async_to_raw_response_wrapper(
            plan_templates.list,
        )
        self.delete = async_to_raw_response_wrapper(
            plan_templates.delete,
        )


class PlanTemplatesResourceWithStreamingResponse:
    def __init__(self, plan_templates: PlanTemplatesResource) -> None:
        self._plan_templates = plan_templates

        self.create = to_streamed_response_wrapper(
            plan_templates.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            plan_templates.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            plan_templates.update,
        )
        self.list = to_streamed_response_wrapper(
            plan_templates.list,
        )
        self.delete = to_streamed_response_wrapper(
            plan_templates.delete,
        )


class AsyncPlanTemplatesResourceWithStreamingResponse:
    def __init__(self, plan_templates: AsyncPlanTemplatesResource) -> None:
        self._plan_templates = plan_templates

        self.create = async_to_streamed_response_wrapper(
            plan_templates.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            plan_templates.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            plan_templates.update,
        )
        self.list = async_to_streamed_response_wrapper(
            plan_templates.list,
        )
        self.delete = async_to_streamed_response_wrapper(
            plan_templates.delete,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime
from typing_extensions import Literal

from pydantic import Field as FieldInfo

from .._models import BaseModel

__all__ = ["Balance"]


class Balance(BaseModel):
    id: str
    """The UUID of the entity."""

    version: int
    """The version number:

    - **Create:** On initial Create to insert a new entity, the version is set at 1
      in the response.
    - **Update:** On successful Update, the version is incremented by 1 in the
      response.
    """

    account_id: Optional[str] = FieldInfo(alias="accountId", default=None)
    """
    The unique identifier (UUID) for the end customer Account the Balance belongs
    to.
    """

    amount: Optional[float] = None
    """The financial value that the Balance holds."""

    balance_draw_down_description: Optional[str] = FieldInfo(alias="balanceDrawDownDescription", default=None)
    """
    A description for the bill line items for charges drawn-down against the
    Balance.
    """

    code: Optional[str] = None
    """A unique short code assigned to the Balance."""

    consumptions_accounting_product_id: Optional[str] = FieldInfo(alias="consumptionsAccountingProductId", default=None)

    created_by: Optional[str] = FieldInfo(alias="createdBy", default=None)
    """The unique identifier (UUID) for the user who created the Balance."""

    currency: Optional[str] = None
    """The currency code used for the Balance amount. For example: USD, GBP or EUR."""

    description: Optional[str] = None
    """A description of the Balance."""

    dt_created: Optional[datetime] = FieldInfo(alias="dtCreated", default=None)
    """The date and time _(in ISO 8601 format)_ when the Balance was first created."""

    dt_last_modified: Optional[datetime] = FieldInfo(alias="dtLastModified", default=None)
    """The date and time _(in ISO 8601 format)_ when the Balance was last modified."""

    end_date: Optional[datetime] = FieldInfo(alias="endDate", default=None)
    """
    The date _(in ISO 8601 format)_ after which the Balance will no longer be
    active.
    """

    fees_accounting_product_id: Optional[str] = FieldInfo(alias="feesAccountingProductId", default=None)

    last_modified_by: Optional[str] = FieldInfo(alias="lastModifiedBy", default=None)
    """The unique identifier (UUID) for the user who last modified the Balance."""

    line_item_types: Optional[
        List[
            Literal[
                "STANDING_CHARGE", "USAGE", "MINIMUM_SPEND", "COUNTER_RUNNING_TOTAL_CHARGE", "COUNTER_ADJUSTMENT_DEBIT"
            ]
        ]
    ] = FieldInfo(alias="lineItemTypes", default=None)
    """
    A list of line item charge types that can draw-down against the Balance amount
    at billing.
    """

    name: Optional[str] = None
    """The official name of the Balance."""

    overage_description: Optional[str] = FieldInfo(alias="overageDescription", default=None)
    """A description for overage charges."""

    overage_surcharge_percent: Optional[float] = FieldInfo(alias="overageSurchargePercent", default=None)
    """
    The percentage surcharge applied to overage charges _(usage above the Balance)_.
    """

    product_ids: Optional[List[str]] = FieldInfo(alias="productIds", default=None)
    """
    A list of Product IDs whose consumption charges due at billing can be drawn-down
    against the Balance amount.
    """

    rollover_amount: Optional[float] = FieldInfo(alias="rolloverAmount", default=None)
    """
    The maximum amount that can be carried over past the Balance end date and
    draw-down against for billing if there is an unused Balance amount remaining
    when the Balance end date is reached.
    """

    rollover_end_date: Optional[datetime] = FieldInfo(alias="rolloverEndDate", default=None)
    """
    The end date _(in ISO 8601 format)_ for the rollover grace period, which is the
    period that unused Balance amounts can be carried over beyond the specified
    Balance `endDate` and continue to be drawn-down against for billing.
    """

    start_date: Optional[datetime] = FieldInfo(alias="startDate", default=None)
    """The date _(in ISO 8601 format)_ when the Balance becomes active."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from m3ter import M3ter, AsyncM3ter
from m3ter.types import CustomFieldsResponse
from tests.utils import assert_matches_type

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestCustomFields:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_retrieve(self, client: M3ter) -> None:
        custom_field = client.custom_fields.retrieve(
            org_id="orgId",
        )
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    def test_raw_response_retrieve(self, client: M3ter) -> None:
        response = client.custom_fields.with_raw_response.retrieve(
            org_id="orgId",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        custom_field = response.parse()
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    def test_streaming_response_retrieve(self, client: M3ter) -> None:
        with client.custom_fields.with_streaming_response.retrieve(
            org_id="orgId",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            custom_field = response.parse()
            assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve(self, client: M3ter) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `org_id` but received ''"):
            client.custom_fields.with_raw_response.retrieve(
                org_id="",
            )

    @parametrize
    def test_method_update(self, client: M3ter) -> None:
        custom_field = client.custom_fields.update(
            org_id="orgId",
        )
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    def test_method_update_with_all_params(self, client: M3ter) -> None:
        custom_field = client.custom_fields.update(
            org_id="orgId",
            account={"foo": "string"},
            account_plan={"foo": "string"},
            aggregation={"foo": "string"},
            compound_aggregation={"foo": "string"},
            meter={"foo": "string"},
            organization={"foo": "string"},
            plan={"foo": "string"},
            plan_template={"foo": "string"},
            product={"foo": "string"},
            version=0,
        )
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    def test_raw_response_update(self, client: M3ter) -> None:
        response = client.custom_fields.with_raw_response.update(
            org_id="orgId",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        custom_field = response.parse()
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    def test_streaming_response_update(self, client: M3ter) -> None:
        with client.custom_fields.with_streaming_response.update(
            org_id="orgId",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            custom_field = response.parse()
            assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update(self, client: M3ter) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `org_id` but received ''"):
            client.custom_fields.with_raw_response.update(
                org_id="",
            )


class TestAsyncCustomFields:
    parametrize = pytest.mark.parametrize("async_client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    async def test_method_retrieve(self, async_client: AsyncM3ter) -> None:
        custom_field = await async_client.custom_fields.retrieve(
            org_id="orgId",
        )
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncM3ter) -> None:
        response = await async_client.custom_fields.with_raw_response.retrieve(
            org_id="orgId",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        custom_field = await response.parse()
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncM3ter) -> None:
        async with async_client.custom_fields.with_streaming_response.retrieve(
            org_id="orgId",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            custom_field = await response.parse()
            assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncM3ter) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `org_id` but received ''"):
            await async_client.custom_fields.with_raw_response.retrieve(
                org_id="",
            )

    @parametrize
    async def test_method_update(self, async_client: AsyncM3ter) -> None:
        custom_field = await async_client.custom_fields.update(
            org_id="orgId",
        )
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    async def test_method_update_with_all_params(self, async_client: AsyncM3ter) -> None:
        custom_field = await async_client.custom_fields.update(
            org_id="orgId",
            account={"foo": "string"},
            account_plan={"foo": "string"},
            aggregation={"foo": "string"},
            compound_aggregation={"foo": "string"},
            meter={"foo": "string"},
            organization={"foo": "string"},
            plan={"foo": "string"},
            plan_template={"foo": "string"},
            product={"foo": "string"},
            version=0,
        )
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    async def test_raw_response_update(self, async_client: AsyncM3ter) -> None:
        response = await async_client.custom_fields.with_raw_response.update(
            org_id="orgId",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        custom_field = await response.parse()
        assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncM3ter) -> None:
        async with async_client.custom_fields.with_streaming_response.update(
            org_id="orgId",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            custom_field = await response.parse()
            assert_matches_type(CustomFieldsResponse, custom_field, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update(self, async_client: AsyncM3ter) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `org_id` but received ''"):
            await async_client.custom_fields.with_raw_response.update(
                org_id="",
            )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable
from datetime import datetime
from typing_extensions import Literal, Required, Annotated, TypedDict

from .._utils import PropertyInfo
from .shared_params.pricing_band import PricingBand

__all__ = ["PricingCreateParams"]


class PricingCreateParams(TypedDict, total=False):
    org_id: Annotated[str, PropertyInfo(alias="orgId")]

    pricing_bands: Required[Annotated[Iterable[PricingBand], PropertyInfo(alias="pricingBands")]]

    start_date: Required[Annotated[Union[str, datetime], PropertyInfo(alias="startDate", format="iso8601")]]
    """
    The start date _(in ISO-8601 format)_ for when the Pricing starts to be active
    for the Plan of Plan Template._(Required)_
    """

    accounting_product_id: Annotated[str, PropertyInfo(alias="accountingProductId")]
    """
    Optional Product ID this Pricing should be attributed to for accounting purposes
    """

    aggregation_id: Annotated[str, PropertyInfo(alias="aggregationId")]
    """UUID of the Aggregation used to create the Pricing.

    Use this when creating a Pricing for a segmented aggregation.
    """

    code: str
    """Unique short code for the Pricing."""

    compound_aggregation_id: Annotated[str, PropertyInfo(alias="compoundAggregationId")]
    """UUID of the Compound Aggregation used to create the Pricing."""

    cumulative: bool
    """
    Controls whether or not charge rates under a set of pricing bands configured for
    a Pricing are applied according to each separate band or at the highest band
    reached.

    _(Optional)_. The default value is **FALSE**.

    - When TRUE, at billing charge rates are applied according to each separate
      band.

    - When FALSE, at billing charge rates are applied according to highest band
      reached.

    **NOTE:** Use the `cumulative` parameter to create the type of Pricing you
    require. For example, for Tiered Pricing set to **TRUE**; for Volume Pricing,
    set to **FALSE**.
    """

    description: str
    """Displayed on Bill line items."""

    end_date: Annotated[Union[str, datetime], PropertyInfo(alias="endDate", format="iso8601")]
    """
    The end date _(in ISO-8601 format)_ for when the Pricing ceases to be active for
    the Plan or Plan Template.

    _(Optional)_ If not specified, the Pricing remains active indefinitely.
    """

    minimum_spend: Annotated[float, PropertyInfo(alias="minimumSpend")]
    """
    The minimum spend amount per billing cycle for end customer Accounts on a Plan
    to which the Pricing is applied.
    """

    minimum_spend_bill_in_advance: Annotated[bool, PropertyInfo(alias="minimumSpendBillInAdvance")]
    """The default value is **FALSE**.

    - When TRUE, minimum spend is billed at the start of each billing period.

    - When FALSE, minimum spend is billed at the end of each billing period.

    _(Optional)_. Overrides the setting at Organization level for minimum spend
    billing in arrears/in advance.
    """

    minimum_spend_description: Annotated[str, PropertyInfo(alias="minimumSpendDescription")]
    """Minimum spend description _(displayed on the bill line item)_."""

    overage_pricing_bands: Annotated[Iterable[PricingBand], PropertyInfo(alias="overagePricingBands")]
    """
    Specify Prepayment/Balance overage pricing in pricing bands for the case of a
    **Tiered** pricing structure.
    """

    plan_id: Annotated[str, PropertyInfo(alias="planId")]
    """UUID of the Plan the Pricing is created for."""

    plan_template_id: Annotated[str, PropertyInfo(alias="planTemplateId")]
    """UUID of the Plan Template the Pricing is created for."""

    segment: Dict[str, str]
    """
    Specifies the segment value which you are defining a Pricing for using this
    call:

    - For each segment value defined on a Segmented Aggregation you must create a
      separate Pricing and use the appropriate `aggregationId` parameter for the
      call.
    - If you specify a segment value that has not been defined for the Aggregation,
      you'll receive an error.
    - If you've defined segment values for the Aggregation using a single wildcard
      or multiple wildcards, you can create Pricing for these wildcard segment
      values also.

    For more details on creating Pricings for segment values on a Segmented
    Aggregation using this call, together with some examples, see the
    [Using API Call to Create Segmented Pricings](https://www.m3ter.com/docs/guides/plans-and-pricing/pricing-plans/pricing-plans-using-segmented-aggregations#using-api-call-to-create-a-segmented-pricing)
    in our User Documentation.
    """

    tiers_span_plan: Annotated[bool, PropertyInfo(alias="tiersSpanPlan")]
    """The default value is **FALSE**.

    - If TRUE, usage accumulates over the entire period the priced Plan is active
      for the account, and is not reset for pricing band rates at the start of each
      billing period.

    - If FALSE, usage does not accumulate, and is reset for pricing bands at the
      start of each billing period.
    """

    type: Literal["DEBIT", "PRODUCT_CREDIT", "GLOBAL_CREDIT"]
    """- **DEBIT**.

    Default setting. The amount calculated using the Pricing is added to the bill as
    a debit.

    - **PRODUCT_CREDIT**. The amount calculated using the Pricing is added to the
      bill as a credit _(negative amount)_. To prevent negative billing, the bill
      will be capped at the total of other line items for the same Product.

    - **GLOBAL_CREDIT**. The amount calculated using the Pricing is added to the
      bill as a credit _(negative amount)_. To prevent negative billing, the bill
      will be capped at the total of other line items for the entire bill, which
      might include other Products the Account consumes.
    """

    version: int
    """The version number of the entity:

    - **Create entity:** Not valid for initial insertion of new entity - _do not use
      for Create_. On initial Create, version is set at 1 and listed in the
      response.
    - **Update Entity:** On Update, version is required and must match the existing
      version because a check is performed to ensure sequential versioning is
      preserved. Version is incremented by 1 and listed in the response.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["ResourceGroupListContentsParams"]


class ResourceGroupListContentsParams(TypedDict, total=False):
    org_id: Annotated[str, PropertyInfo(alias="orgId")]

    type: Required[str]

    next_token: Annotated[str, PropertyInfo(alias="nextToken")]
    """nextToken for multi page retrievals"""

    page_size: Annotated[int, PropertyInfo(alias="pageSize")]
    """Number of ResourceGroupItems to retrieve per page"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from pydantic import Field as FieldInfo

from .._models import BaseModel

__all__ = ["CounterResponse"]


class CounterResponse(BaseModel):
    id: str
    """The UUID of the entity."""

    version: int
    """The version number:

    - **Create:** On initial Create to insert a new entity, the version is set at 1
      in the response.
    - **Update:** On successful Update, the version is incremented by 1 in the
      response.
    """

    code: Optional[str] = None
    """Code of the Counter. A unique short code to identify the Counter."""

    created_by: Optional[str] = FieldInfo(alias="createdBy", default=None)
    """The ID of the user who created this item."""

    dt_created: Optional[datetime] = FieldInfo(alias="dtCreated", default=None)
    """The DateTime when this item was created _(in ISO-8601 format)_."""

    dt_last_modified: Optional[datetime] = FieldInfo(alias="dtLastModified", default=None)
    """The DateTime when this item was last modified _(in ISO-8601 format)_."""

    last_modified_by: Optional[str] = FieldInfo(alias="lastModifiedBy", default=None)
    """The ID of the user who last modified this item."""

    name: Optional[str] = None
    """Name of the Counter."""

    product_id: Optional[str] = FieldInfo(alias="productId", default=None)
    """UUID of the product the Counter belongs to.

    _(Optional)_ - if no `productId` is returned, the Counter is Global. A Global
    Counter can be used to price Plans or Plan Templates belonging to any Product.
    """

    unit: Optional[str] = None
    """
    Label for units shown on Bill line items, and indicating to customers what they
    are being charged for.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .invitation_response import InvitationResponse as InvitationResponse
from .invitation_list_params import InvitationListParams as InvitationListParams
from .invitation_create_params import InvitationCreateParams as InvitationCreateParams



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List, Union
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from ..._utils import PropertyInfo

__all__ = ["InvitationCreateParams"]


class InvitationCreateParams(TypedDict, total=False):
    org_id: Annotated[str, PropertyInfo(alias="orgId")]

    email: Required[str]

    first_name: Required[Annotated[str, PropertyInfo(alias="firstName")]]

    last_name: Required[Annotated[str, PropertyInfo(alias="lastName")]]

    contact_number: Annotated[str, PropertyInfo(alias="contactNumber")]

    dt_end_access: Annotated[Union[str, datetime], PropertyInfo(alias="dtEndAccess", format="iso8601")]
    """The date when access will end for the user _(in ISO-8601 format)_.

    Leave blank for no end date, which gives the user permanent access.
    """

    dt_expiry: Annotated[Union[str, datetime], PropertyInfo(alias="dtExpiry", format="iso8601")]
    """The date when the invite expires _(in ISO-8601 format)_.

    After this date the invited user can no longer accept the invite. By default,
    any invite is valid for 30 days from the date the invite is sent.
    """

    m3ter_user: Annotated[bool, PropertyInfo(alias="m3terUser")]

    permission_policy_ids: Annotated[List[str], PropertyInfo(alias="permissionPolicyIds")]
    """The IDs of the permission policies the invited user has been assigned.

    This controls the access rights and privileges that this user will have when
    working in the m3ter Organization.
    """

    version: int



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List, Optional
from typing_extensions import Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["CommitmentListParams"]


class CommitmentListParams(TypedDict, total=False):
    org_id: Annotated[str, PropertyInfo(alias="orgId")]

    account_id: Annotated[str, PropertyInfo(alias="accountId")]
    """The unique identifier (UUID) for the Account.

    This parameter helps filter the Commitments related to a specific end-customer
    Account.
    """

    contract_id: Annotated[Optional[str], PropertyInfo(alias="contractId")]

    date: str
    """
    A date _(in ISO-8601 format)_ to filter Commitments which are active on this
    specific date.
    """

    end_date_end: Annotated[str, PropertyInfo(alias="endDateEnd")]
    """A date _(in ISO-8601 format)_ used to filter Commitments.

    Only Commitments with end dates before this date will be included.
    """

    end_date_start: Annotated[str, PropertyInfo(alias="endDateStart")]
    """A date _(in ISO-8601 format)_ used to filter Commitments.

    Only Commitments with end dates on or after this date will be included.
    """

    ids: List[str]
    """A list of unique identifiers (UUIDs) for the Commitments to retrieve.

    Use this to fetch specific Commitments in a single request.
    """

    next_token: Annotated[str, PropertyInfo(alias="nextToken")]
    """The `nextToken` for multi-page retrievals.

    It is used to fetch the next page of Commitments in a paginated list.
    """

    page_size: Annotated[int, PropertyInfo(alias="pageSize")]
    """Specifies the maximum number of Commitments to retrieve per page."""

    product_id: Annotated[str, PropertyInfo(alias="productId")]
    """The unique identifier (UUID) for the Product.

    This parameter helps filter the Commitments related to a specific Product.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from pydantic import Field as FieldInfo

from .._models import BaseModel

__all__ = ["AdHocResponse"]


class AdHocResponse(BaseModel):
    job_id: Optional[str] = FieldInfo(alias="jobId", default=None)
    """The id of the job"""
